use std::{
    collections::BTreeMap,
    fmt,
};
use serde::{
    Serialize, Deserialize,
    //Serialize, Serializer, Deserialize, Deserializer,
    //ser::SerializeMap,
    //de::{ Visitor, MapAccess, },
};
use ron::{
    ser::PrettyConfig,
    extensions::Extensions,
    Options,
};

use crate::{
    route::*,
    CowStr,
};


#[derive(Debug)]
pub enum ConfigParseError {
    RecursiveForward(String),
    DuplicateKeys(Vec<String>),
}

impl std::error::Error for ConfigParseError {}

impl fmt::Display for ConfigParseError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::RecursiveForward(addr) => write!(f, "Forwarding to the address {addr} would be recursive"),
            Self::DuplicateKeys(dups) => {
                write!(f, "There are duplicate entries for these host names/URLs:")?;
                for key in dups {
                    write!(f, " {key}")?;
                }
                Ok(())
            },
        }
    }
}


#[derive(Debug)]
pub struct MultiKeyMap<K: Ord + Clone, V> {
    keys: BTreeMap<K, usize>,
    vals: Vec<V>,
}

impl<K: Ord + Clone, V> MultiKeyMap<K, V> {
    pub fn new() -> Self {
        Self {
            keys: BTreeMap::new(),
            vals: Vec::new(),
        }
    }

    pub fn insert<'k>(&mut self, keys: &'k [K], val: V) -> Result<(), Vec<&'k K>> {
        let index = self.vals.len();
        self.vals.push(val);

        let mut duplicates: Vec<&K> = Vec::new();
        for key in keys {
            if self.keys.contains_key(key) {
                duplicates.push(key);
            } else {
                self.keys.insert(key.clone(), index);
                assert!(self.keys.get(key).is_some());
            }
        }

        if duplicates.is_empty() {
            Ok(())
        } else {
            Err(duplicates)
        }
    }

    pub fn get(&self, key: &K) -> Option<&V> {
        self.keys.get(key)
            .map(|i| &self.vals[*i])
    }
}


#[derive(Debug, Serialize, Deserialize)]
pub struct RawConfig {
    pub bind: String,
    pub routes: BTreeMap<Option<Vec<CowStr>>, BTreeMap<Option<Vec<CowStr>>, Route>>,
}

impl RawConfig {
    pub fn from_str(s: &str) -> ron::error::SpannedResult<Self> {
        let ron = Options::default().with_default_extension(Extensions::all());
        ron.from_str(s)
    }

    pub fn to_string(&self) -> ron::error::Result<String> {
        let ron = Options::default().with_default_extension(Extensions::all());
        ron.to_string_pretty(self, PrettyConfig::new())
    }

    pub fn into_config(self) -> Result<Config, ConfigParseError> {
        let mut duplicates: Vec<String> = Vec::new();
        let mut hosts = MultiKeyMap::new();

        for (host_paths, host) in self.routes.into_iter() {
            let mut routes = MultiKeyMap::new();
            for (route_paths, route) in host.into_iter() {
                let mapped_route_paths: Vec<Option<CowStr>> = route_paths.map(|rp| rp.into_iter().map(|r| Some(r)).collect()).unwrap_or_else(|| vec![None]);
                if let Route::Forward(fwd) = &route {
                    if fwd == &self.bind {
                        return Err(ConfigParseError::RecursiveForward(self.bind));
                    }
                }
                routes.insert(&mapped_route_paths, route).unwrap_or_else(|e| {
                    let dups: Vec<String> = e.into_iter().filter_map(|r| r.as_ref()).map(|r| r.to_string()).collect();
                    duplicates.extend_from_slice(&dups);
                });
            }
            let mapped_host_paths: Vec<Option<CowStr>> = host_paths.map(|hp| hp.into_iter().map(|h| Some(h)).collect()).unwrap_or_else(|| vec![None]);
            hosts.insert(&mapped_host_paths, routes).unwrap_or_else(|e| {
                let dups: Vec<String> = e.into_iter().filter_map(|r| r.as_ref()).map(|r| r.to_string()).collect();
                duplicates.extend_from_slice(&dups);
            });
        }

        if !duplicates.is_empty() {
            duplicates.sort();
            // Ironic
            duplicates.dedup();
            Err(ConfigParseError::DuplicateKeys(duplicates))
        } else {
            Ok(Config {
                bind: self.bind,
                routes: hosts,
            })
        }
    }
}

impl Default for RawConfig {
    fn default() -> Self {
        Self {
            bind: "0.0.0.0:8080".into(),
            routes: {
                let mut hosts = BTreeMap::new();
                hosts.insert(None, {
                    let mut routes = BTreeMap::new();
                    routes.insert(Some(vec!["/".into()]), Route::Hardcode(HardcodeRoute { content: r#"<!DOCTYPE HTML><html><body><a href="/page.ron">page.ron</a><br><a href="/fallback">Fallback</a></body></html>"#.into(), mime: "text/html".into(), ..Default::default() }));
                    routes.insert(Some(vec!["/page.ron".into()]), Route::File(FileRoute { path: "page.ron".into(), mime: "application/ron".into(), ..Default::default() }));
                    routes.insert(None, Route::Hardcode(HardcodeRoute { content: "Fallback page".into(), mime: "text/plain".into(), code: Some("404 Not Found".into()), ..Default::default() }));
                    routes
                });
                hosts.insert(Some(vec!["mysite.com".into(), "mysite.net".into()]), {
                    let mut routes = BTreeMap::new();
                    routes.insert(None, Route::Hardcode(HardcodeRoute { content: "Different site depending on Host header".into(), mime: "text/plain".into(), ..Default::default() }));
                    routes
                });
                hosts
            },
        }
    }
}


#[derive(Debug)]
pub struct Config {
    pub bind: String,
    pub routes: MultiKeyMap<Option<CowStr>, MultiKeyMap<Option<CowStr>, Route>>,
}
