use std::{
    iter,
};
use tokio::{
    net::TcpStream,
    io::{ AsyncWriteExt, },
};

use crate::CowStr;


fn find_r_n(buf: &[u8]) -> Result<usize, ()> {
    let mut iter = buf.iter().enumerate().take(128);
    let mut prev_byte = b'\0';
    while let Some((index, byte)) = iter.next() {
        if prev_byte == b'\r' && *byte == b'\n' {
            return Ok(index + 1);
        }
        prev_byte = *byte;
    }
    Err(())
}


#[derive(Debug)]
pub struct RequestHeader<'h> {
    buf: &'h [u8],
    headers_start: usize,
}

impl<'h> RequestHeader<'h> {
    pub fn new(buf: &'h [u8]) -> Self {
        Self {
            buf,
            headers_start: 0,
        }
    }

    pub fn url(&mut self) -> Option<&'h str> {
        let next_line = find_r_n(&self.buf).ok()?;
        let bytes = &self.buf[..next_line];

        self.headers_start = next_line;
        let line = std::str::from_utf8(bytes).ok()?;

        let mut req_parts = line.split_ascii_whitespace();
        // who cares about the method & version amirite
        req_parts.next()?;
        req_parts.next()
    }

    pub fn headers(&'h self, filter: &'h str) -> Headers<'h> {
        Headers {
            buf: &self.buf[self.headers_start..],
            filter,
            cursor: 0,
        }
    }
}


pub struct Headers<'h> {
    buf: &'h [u8],
    filter: &'h str,
    cursor: usize,
}

impl<'h> iter::Iterator for Headers<'h> {
    type Item = &'h str;

    fn next(&mut self) -> Option<Self::Item> {
        loop {
            let next_line = find_r_n(&self.buf[self.cursor..]).ok()?;
            let bytes = &self.buf[self.cursor..self.cursor + next_line];
            if bytes == b"\r\n" {
                return None;
            }
            self.cursor += next_line;

            let line = std::str::from_utf8(bytes).ok()?;
            let (key, mut value) = line.split_once(":")?;
            let key_lower = key.to_lowercase();
            if key_lower == self.filter {
                value = value.trim_start().strip_suffix("\r\n")?;
                return Some(value);
            }
        }
    }
}


pub struct ResponseHeader {
    pub code: CowStr,
    pub headers: Vec<(CowStr, CowStr)>,
}

impl ResponseHeader {
    pub fn with_code(code: CowStr) -> Self {
        Self {
            code,
            headers: Vec::new(),
        }
    }

    pub async fn write_to(&self, socket: &mut TcpStream) -> tokio::io::Result<()> {
        socket.write_all(format!("HTTP/1.1 {}\r\n", self.code).as_bytes()).await?;
        for (key, value) in &self.headers {
            socket.write_all(format!("{key}: {value}\r\n").as_bytes()).await?;
        }
        socket.write_all(b"\r\n").await?;
        socket.flush().await?;
        Ok(())
    }
}
