use std::{
    borrow::Cow,
    sync::Arc,
};
use chrono::{ TimeDelta, Local, };
use tokio::{
    net::{ TcpListener, },
};

mod config;
mod route;
pub mod http;
use config::*;
use http::*;


/// Log time if connection received 10 mins after last log
pub const LOG_TIME_EVERY: TimeDelta = TimeDelta::minutes(10);

type CowStr = Cow<'static, str>;


pub async fn run() -> Result<(), Box<dyn std::error::Error>> {
    let raw_config = {
        match std::fs::read_to_string("page.ron") {
            Ok(raw_config) => {
                RawConfig::from_str(&raw_config).expect("Error parsing config")
            },
            Err(_) => {
                eprintln!("Could not open config file, creating new file");
                let raw_config = RawConfig::default();
                std::fs::write(
                    "page.ron",
                    raw_config.to_string().expect("Error serializing default config")
                ).expect("Error creating default config file");
                raw_config
            }
        }
    };

    let config = Arc::new(raw_config.into_config()?);

    let listener = TcpListener::bind(&config.bind).await?;

    let mut last_log = Local::now();
    eprintln!("Started at {} -- Version {}", last_log.format("%F %T"), env!("CARGO_PKG_VERSION"));

    loop {
        let (mut socket, addr) = listener.accept().await?;

        let now = Local::now();
        if now - last_log > LOG_TIME_EVERY {
            last_log = now;
            eprintln!("-- {} --", now.format("%F %T"));
        }

        let config = config.clone();
        tokio::spawn(async move {
            if socket.readable().await.is_err() {
                return;
            }

            let mut data = [0u8; 8192];
            let n = socket.peek(&mut data).await;
            let Ok(bytes_read) = n else {
                return;
            };

            let mut header = RequestHeader::new(&data[..bytes_read]);

            let url = header.url().and_then(|h| {
                let mut s = h;
                if let Some(n) = s.find("?") {
                    s = &s[..n];
                }
                if let Some(n) = s.find("#") {
                    s = &s[..n];
                }
                //Cow::Borrowed(s)
                percent_encoding::percent_decode(&s.as_bytes()).decode_utf8().ok()
            });

            let host = header.headers("host").next();

            eprintln!(
                "{}\t{}\t{}",
                addr,
                host.as_ref().map(|h| &**h).unwrap_or("~"),
                url.as_ref().map(|h| &**h).unwrap_or("~"),
            );

            let host_route = config.routes.get(&host.map(|h| Cow::Borrowed(h)))
                .or_else(|| config.routes.get(&None));

            let route = host_route.and_then(|hr| {
                hr.get(&url)
                    .or_else(|| hr.get(&None))
            });

            match route {
                Some(r) => {
                    r.get(&mut socket).await
                        .unwrap_or_else(|e| eprintln!("\t{e}"));
                },
                None => {
                    ResponseHeader::with_code("404 Not Found".into()).write_to(&mut socket).await
                        .unwrap_or_else(|e| eprintln!("\t{e}"));
                },
            }
        });
    }
}
