use serde::{ Serialize, Deserialize, };
use tokio::{
    net::TcpStream,
    io::{ AsyncWriteExt, },
    fs::{ File, },
};

use crate::{
    http::*,
    CowStr,
};


// serde moment
fn is_false(b: &bool) -> bool {
    !b
}


#[derive(Debug, Serialize, Deserialize)]
pub struct HardcodeRoute {
        pub content: CowStr,
        pub mime: CowStr,
        #[serde(default, skip_serializing_if = "Option::is_none")]
        pub code: Option<CowStr>,
        #[serde(default, skip_serializing_if = "Vec::is_empty")]
        pub headers: Vec<(CowStr, CowStr)>,
}

impl Default for HardcodeRoute {
    fn default() -> Self {
        Self {
            content: "".into(),
            mime: "text/html".into(),
            code: None,
            headers: Vec::new(),
        }
    }
}


#[derive(Debug, Serialize, Deserialize)]
pub struct FileRoute {
        pub path: CowStr,
        pub mime: CowStr,
        #[serde(default, skip_serializing_if = "Option::is_none")]
        pub code: Option<CowStr>,
        #[serde(default, skip_serializing_if = "Vec::is_empty")]
        pub headers: Vec<(CowStr, CowStr)>,
        #[serde(default, skip_serializing_if = "is_false")]
        pub gzip: bool,
        //#[serde(default, skip_serializing_if = "is_false")]
        //pub chunk: bool,
}

impl Default for FileRoute {
    fn default() -> Self {
        Self {
            path: "".into(),
            mime: "text/html".into(),
            code: None,
            headers: Vec::new(),
            gzip: false,
            //chunk: false,
        }
    }
}


#[derive(Debug, Serialize, Deserialize)]
pub enum Route {
    Hardcode(HardcodeRoute),
    File(FileRoute),
    Forward(CowStr),
}

impl Route {
    pub async fn get(&self, socket: &mut TcpStream) -> Result<(), tokio::io::Error> {
        match self {
            Self::Hardcode(hardcode_route) => {
                drain_socket(socket).await;

                let mut response = ResponseHeader::with_code(hardcode_route.code.clone().unwrap_or("200 OK".into()));

                response.headers.extend_from_slice(&hardcode_route.headers);
                response.headers.push(("Content-Type".into(), hardcode_route.mime.clone()));
                response.headers.push(("Content-Length".into(), format!("{}", hardcode_route.content.len()).into()));
                response.write_to(socket).await?;

                socket.write_all(hardcode_route.content.as_bytes()).await?;
                let _ = socket.shutdown().await;
            },

            Self::File(file_route) => {
                drain_socket(socket).await;

                let mut file = match File::open(&*file_route.path).await {
                    Ok(f) => f,
                    Err(e) => {
                        eprintln!("Could not open {} for reading: {e}", file_route.path);
                        ResponseHeader::with_code("500 Internal Server Error".into())
                            .write_to(socket).await?;
                        return Err(e);
                    },
                };
                let file_len = match file.metadata().await {
                    Ok(m) => m.len(),
                    Err(e) => {
                        eprintln!("Could not get metadata of {}: {e}", file_route.path);
                        ResponseHeader::with_code("500 Internal Server Error".into())
                            .write_to(socket).await?;
                        return Err(e);
                    },
                };

                let mut response = ResponseHeader::with_code(file_route.code.clone().unwrap_or("200 OK".into()));

                if file_route.gzip {
                    response.headers.push(("Content-Encoding".into(), "gzip".into()));
                    response.headers.push(("Vary".into(), "Accept-Encoding".into()));
                }
                response.headers.extend_from_slice(&file_route.headers);
                response.headers.push(("Content-Type".into(), file_route.mime.clone()));
                response.headers.push(("Content-Length".into(), format!("{file_len}").into()));
                response.write_to(socket).await?;

                tokio::io::copy(&mut file, socket).await?;

                let _ = socket.shutdown().await;
            },

            Self::Forward(addr) => {
                let mut forward = TcpStream::connect(&**addr).await?;

                {
                    let mut buf = [0; 8192];
                    let mut tries = 0;
                    loop {
                        match socket.try_read(&mut buf) {
                            Ok(read @ 1..) => {
                                tries = 0;
                                forward.write_all(&buf[..read]).await?;
                            },
                            _ => {
                                tries += 1;
                                if tries >= 4 {
                                    break;
                                }
                            },
                        }
                    }
                }
                {
                    forward.readable().await?;
                    let mut buf = [0; 8192];
                    let mut tries = 0;
                    loop {
                        match forward.try_read(&mut buf) {
                            Ok(read @ 1..) => {
                                tries = 0;
                                socket.write_all(&buf[..read]).await?;
                            },
                            _ => {
                                tries += 1;
                                if tries >= 4 {
                                    break;
                                }
                            },
                        }
                    }
                }

                let _ = socket.shutdown().await;
                let _ = forward.shutdown().await;
            },
        }
        Ok(())
    }
}


async fn drain_socket(socket: &mut TcpStream) {
    let mut _sink = [0; 8192];
    let mut tries = 0;
    loop {
        match socket.try_read(&mut _sink) {
            Ok(1..) => tries = 0,
            _ => {
                tries += 1;
                if tries >= 4 {
                    break;
                }
            },
        }
    }
}
